library firestore_service;

import 'dart:async';
import 'dart:convert';
import 'dart:html' as dom;
import 'dart:math';
import 'package:angular/angular.dart';
import 'package:bokain_models/bokain_models.dart';
//import 'package:fo_components/fo_components.dart';
import 'package:firebase/firebase.dart' as fb;
import 'package:firebase/firestore.dart';

part 'booking_service.dart';
part 'chat_service.dart';
part 'consultation_service.dart';
part 'customer_service.dart';
part 'day_service.dart';
part 'ingredient_service.dart';
part 'issue_service.dart';
part 'journal_service.dart';
part 'product_category_service.dart';
part 'product_routine_service.dart';
part 'product_service.dart';
part 'room_service.dart';
part 'salon_service.dart';
part 'service_addon_service.dart';
part 'service_service.dart';
part 'settings_service.dart';
part 'user_service.dart';
part 'webshop_content_service.dart';

abstract class FirestoreService<T> {
  FirestoreService([String path, this._uniqueFields = const []]) {
    setPath(path);
  }

  void setPath(String path) {
    cancelStreaming();
    _collection = null;
    _path = path;
    if (_path != null) {
      _collection = fb.firestore().collection(_path);
    }
  }

  Future<String> push(T model) async {
    if (_path == null || _collection == null)
      throw new StateError('Path is not set');
    if ((model as EditableModel).id != null)
      throw new StateError('This model already exists');

    _loading = true;

    (model as EditableModel)
      ..created = new DateTime.now()
      ..added_by = currentUser.id;

    final data = json.decode(json.encode(model))
      ..removeWhere((key, value) => value == null);

    final ref = await _collection.add(data);
    (model as EditableModel).id = ref.id;
    await _collection.doc(ref.id).update(data: {'id': ref.id});

    _loading = false;
    return ref.id;
  }

  Future set(T model) async {
    if (_path == null || _collection == null)
      throw new StateError('Path is not set');
    if (model == null) throw new ArgumentError.notNull('model');

    if ((model as EditableModel).id == null)
      throw new StateError('model id cannot be null');
    _loading = true;
    final Map<String, dynamic> data = json.decode(json.encode(model))
      ..removeWhere((key, value) => value == null);

    await _collection.doc((model as EditableModel).id).update(data: data);
    _loading = false;
  }

  Future patch(String id, Map<String, dynamic> data) async {
    if (_path == null || _collection == null)
      throw new StateError('Path is not set');
    if (id == null) throw new ArgumentError.notNull('id');
    if (data == null) throw new ArgumentError.notNull('data');

    _loading = true;

    data.removeWhere((key, value) => value == null);
    await _collection.doc(id).update(data: data);
    _loading = false;
  }

  void cancelStreaming() {
    _snapshotSub?.cancel();
    _snapshotSub = null;
  }

  Future<List<T>> search(String column, String keywords,
      {int limit = 100, bool clear = true}) async {
    if (keywords == null || keywords.isEmpty) {
      return [];
    }

    _loading = true;
    if (clear) _cachedModels.clear();
    final strLength = keywords.length;
    final strFrontCode = keywords.substring(0, strLength - 1);
    final strEndCode = keywords.substring(strLength - 1, strLength);
    final endCode =
        strFrontCode + new String.fromCharCode(strEndCode.codeUnitAt(0) + 1);
    final query = _collection
        .orderBy(column)
        .where(column, '>=', keywords)
        .where(column, '<', endCode)
        .limit(limit);
    final snapshot = await query.get();
    try {
      snapshot.docs.forEach(_onModelAdded);
    } finally {
      _loading = false;
    }

    return _cachedModels.values.toList();
  }

  Future<T> fetch(String id, {bool force: true, bool cache: true}) async {
    if (_path == null || _collection == null)
      throw new StateError('Path is not set');
    if (id == null) throw new ArgumentError.notNull('id');
    _loading = true;

    T model;

    if (force || !_cachedModels.containsKey(id)) {
      final documentSnapshot = await _collection.doc(id).get();
      if (cache)
        model = _onModelAdded(documentSnapshot);
      else {
        model = createModelInstance(documentSnapshot.data());
        (model as EditableModel).id = documentSnapshot.id;
      }
    } else
      model = _cachedModels[id];

    _loading = false;
    return model;
  }

  T get(String id) => _cachedModels.containsKey(id) ? _cachedModels[id] : null;

  List<T> getMany([Iterable ids]) {
    if (_path == null || _collection == null)
      throw new StateError('Path is not set');
    if (ids == null)
      return _cachedModels.values.toList();
    else {
      return _cachedModels.keys
          .where(ids.contains)
          .map((id) => _cachedModels[id])
          .toList();
    }
  }

  Future<List<T>> fetchMany(
      {String orderBy,
      String orderDir,
      int limit,
      String whereField,
      String whereOp: '==',
      Object whereValue,
      Object whereRangeStart,
      Object whereRangeEnd,
      bool clear = true}) async {
    if (_path == null || _collection == null)
      throw new StateError('Path is not set');
    if (streaming)
      throw new StateError('$_path service is already streaming data');

    _loading = true;
    if (clear) _cachedModels.clear();

    final query = _constructQuery(orderBy, orderDir, limit, whereField, whereOp,
        whereValue, whereRangeStart, whereRangeEnd);
    final snapshot = await query.get();

    try {
      snapshot.docs.forEach(_onModelAdded);
    } finally {
      _loading = false;
    }

    return _cachedModels.values.toList();
  }

  Future remove(String id) async {
    if (_path == null || _collection == null)
      throw new StateError('Path is not set');
    if (id == null) throw new ArgumentError.notNull('id');
    _loading = true;
    await _collection.doc(id).delete();
    _loading = false;
  }

  Stream<T> stream(String id) {
    if (_path == null || _collection == null || id == null)
      throw new StateError('Path is not set');
    if (streaming)
      throw new StateError(
          'this instance of $_path service is already streaming data');
    print('STREAMING $_path/$id');

    _snapshotSub = _collection.doc(id).onSnapshot.listen(_onModelAdded);

    return _onModelAddedController.stream;
  }

  void streamMany(
      {String orderBy,
      String orderDir,
      int limit,
      String whereField,
      String whereOp = '==',
      Object whereValue,
      Object whereRangeStart,
      Object whereRangeEnd,
      bool clear = true}) {
    if (_path == null || _collection == null)
      throw new StateError('Path is not set');
    if (streaming)
      throw new StateError('$_path service is already streaming data');

    print('STREAMING $_path');
    _loading = true;
    new Future.delayed(const Duration(milliseconds: 300))
        .then((_) => _loading = false);

    if (clear) _cachedModels.clear();

    void snapshotCallback(QuerySnapshot snapshot) {
      for (final change in snapshot.docChanges) {
        if (change.type == 'added' || change.type == 'modified') {
          _onModelAdded(change.doc);
        } else if (change.type == 'removed') _onModelRemoved(change.doc);
      }

      _onSnapshotController.add(_cachedModels);
    }

    final query = _constructQuery(orderBy, orderDir, limit, whereField, whereOp,
        whereValue, whereRangeStart, whereRangeEnd);
    _snapshotSub = query.onSnapshot.listen(snapshotCallback);
  }

  Query _constructQuery(
      String orderBy,
      String orderDir,
      int limit,
      String whereField,
      String whereOp,
      Object whereValue,
      Object whereRangeStart,
      Object whereRangeEnd) {
    Query query = _collection;
    if (orderBy != null && orderDir != null)
      query = query.orderBy(orderBy, orderDir);
    if (whereField != null && whereOp != null && whereValue != null)
      query = query.where(whereField, whereOp, whereValue);
    else if (whereField != null &&
        whereRangeStart != null &&
        whereRangeEnd != null)
      query = query.orderBy(whereField).startAt(
          fieldValues: [whereRangeStart]).endAt(fieldValues: [whereRangeEnd]);
    if (limit != null) query = query.limit(limit);

    return query;
  }

  Map<String, T> get cachedModels => _cachedModels;
  bool get loading => _loading;
  bool get streaming => _snapshotSub != null;
  Stream<T> get onModelAdded => _onModelAddedController.stream;
  Stream<String> get onModelRemoved => _onModelRemovedController.stream;
  Stream<Map<String, T>> get onSnapshot => _onSnapshotController?.stream;
  List<String> get uniqueFields => _uniqueFields;
  String get path => _path;

  T createModelInstance(Map<String, dynamic> data);

  T _onModelAdded(DocumentSnapshot document) {
    if (document == null) throw new ArgumentError.notNull('document');

    final data = document.data();
    final id = document.id;

    if (data == null) throw new Exception('document contains no data');
    if (id == null) throw new Exception('document id is null');

    final model = createModelInstance(data);
    (model as EditableModel).id = id;
    _cachedModels[id] = model;

    _loading = true;
    new Future.delayed(const Duration(milliseconds: 100))
        .then((_) => _loading = false);

    _onModelAddedController.add(model);
    return model;
  }

  void _onModelRemoved(DocumentSnapshot document) {
    _cachedModels.remove(document.id);
    print('model ${document.id} removed');
    _onModelRemovedController.add(document.id);
  }

  final Map<String, T> _cachedModels = {};
  String _path;
  CollectionReference _collection;

  StreamSubscription _snapshotSub;
  final StreamController<T> _onModelAddedController =
      new StreamController.broadcast();
  final StreamController<String> _onModelRemovedController =
      new StreamController.broadcast();
  final StreamController<Map<String, T>> _onSnapshotController =
      new StreamController.broadcast();
  bool _loading = false;
  final List<String> _uniqueFields;

  static User currentUser;
}

List<T> filterModelList<T>(List<T> input,
    {List<String> ids,
    String orderBy,
    String orderDir,
    int limit,
    String whereField,
    String whereOp: '==',
    Object whereValue}) {
  final output = input.toList();

  if (ids != null)
    // ignore: avoid_as
    output.removeWhere((model) => !ids.contains((model as EditableModel).id));

  /// orderBy
  if (orderBy != null && orderDir != null) {
    output.sort((t1, t2) {
      /**
       * TODO: proper object sorting (numbers, dates etc)
       */

      final d1 = (t1 as EditableModel).toJson();
      final d2 = (t2 as EditableModel).toJson();

      if (orderDir.toLowerCase() == 'asc')
        return d1[orderBy].toString().compareTo(d2[orderBy].toString());
      else if (orderDir.toLowerCase() == 'desc')
        return d2[orderBy].toString().compareTo(d1[orderBy].toString());
    });
  }

  /// Filter on where
  if (whereField != null && whereOp != null) {
    switch (whereOp) {
      case '==':
        output.removeWhere(
            (t) => (t as EditableModel).toJson()[whereField] != whereValue);
        break;

      case '!=':
        output.removeWhere(
            (t) => (t as EditableModel).toJson()[whereField] == whereValue);
        break;

      case '<':
        if (whereValue == null)
          output.removeWhere(
              (t) => (t as EditableModel).toJson()[whereField] >= whereValue);
        break;

      case '<=':
        output.removeWhere(
            (t) => (t as EditableModel).toJson()[whereField] > whereValue);
        break;

      case '>':
        if (whereValue == null)
          output.removeWhere(
              (t) => (t as EditableModel).toJson()[whereField] <= whereValue);
        break;

      case '>=':
        if (whereValue == null)
          output.removeWhere(
              (t) => (t as EditableModel).toJson()[whereField] < whereValue);
        break;

      default:
        throw new StateError('Invalid whereOp: $whereOp');
        break;
    }
  }

  if (limit != null && output.length > limit)
    output.removeRange(limit, output.length);

  return output;
}

Map<String, T> filterModelMap<T>(Map<String, T> input,
    {List<String> ids,
    String orderBy,
    String orderDir,
    int limit,
    String whereField,
    String whereOp: '==',
    Object whereValue}) {
  final output = <String, T>{};
  final filtered = filterModelList(input.values.toList(),
      ids: ids,
      orderBy: orderBy,
      orderDir: orderDir,
      limit: limit,
      whereField: whereField,
      whereOp: whereOp,
      whereValue: whereValue);
  for (var t in filtered) {
    output[(t as EditableModel).id] = t;
  }
  return output;
}
